unit bootloader;
{    This file is part of the AVR Bootloader by Arne Krüger for the AVRootloader by Hagen Re.

    AVR Bootloader is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
	any later version.

    AVR Bootloader is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with AVR Bootloader.  If not, see <http://www.gnu.org/licenses/>.
	
	This file is programmed by Arne Krüger news.kruegerarne@web.de
}
{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils,synaser,Forms,InputFile,EpikTimer;

const PAGE_SIZE = 64;
type

    TBootInfo = record
      Hello:string;
      Sign0:Byte;
      Sign1:Byte;
      Sign2:Byte;
      version:Byte;
      Bootpages:byte;
      FlashSize:integer;
      Crypt:boolean;
    end;

    TBootloaderStatus = (bsIDLE=0,bsConnecting=$02,bsConnected=$12);

    TBootloaderError = (beNone=0,beTimeOut,beSuccessPlainStyle=$30,beSuccessCryptStyle=$37, beVerify=$C0,beCommand=$C1,beCRC=$C2,beBounds=$C3,beCrypt=$C4,beProg=$C5,beVersion=$C6);

    TBootloaderCMDStatus = (
        csIDlE = $00,
        csKeepAlive = $10,
        csProg_AddressSetted=$20,csProg_AddressOK=$21,csProg_StartNextBuffer=$22,csProg_BufferFilled=$23,csProg_BufferOK=$24,
          csProg_WritePageSent=$25,csProg_WritePageOK=$26,csProg_EraseSent=$27,csProg_EraseOK=$28,csProg_Ready=$29,
          csCrypProg_Start = $2A, csCrypProg_StartNextBuffer = $2B, csCryptProg_CMDSent=$2C,csProgError=$2D
        );
    TBootloaderCMD = (bcNone,bcIDLE,bcKeepAlive,bcProgram);

    TStatusEvent = procedure(Status: TBootloaderStatus; CMDStatus: TBootloaderCMDStatus) of Object;
    TErrorStringEvent = procedure(StatusMsg: string;Error: TBootloaderError) of Object;
    TErrorEvent = procedure(Error: TBootloaderError; Status: TBootloaderStatus; CMDStatus: TBootloaderCMDStatus) of Object;
    TStatusStringEvent = procedure(StatusMsg: string;Status: TBootloaderStatus; const ApplyNewLine:boolean) of Object;
    TBootloaderProgressEvent = procedure(BytesTotal:integer;BytesTransfered:integer) of Object;
    TRXEvent = procedure(RxByte: Byte) of Object;
    TBootInfoEvent = procedure(BootInfo:TBootInfo) of Object;

    TNewPageEvent = procedure(var address:integer;var Buffer: array of Byte; var Length:integer; const PageSize:integer; var FileType:TFileType) of Object;

    { TBootloader }

    { TBootloaderLL }

    { TBootloaderThread }



    TBootloaderThread = class(TThread)
      private

        BootInfo_:TBootInfo;
        ser:TBlockserial;
        fEpikTimer : TEpikTimer;
        fTimerData:TimerData;
        fTimerElapsed_:Extended;

        fBootloaderShouldStatus_ : TBootloaderStatus;
        fBootloaderIsStatus_ : TBootloaderStatus;
        fBootloaderIsStatus  : TBootloaderStatus;
        fBootloaderCommandStatus_:TBootloaderCMDStatus;
        fBootloaderCommand : TBootloaderCMD;
        fBootloaderCommand_ : TBootloaderCMD;

        fTXBuffer_: array[0..1023] of Byte;
        fTXBufferLength_:integer;
        fRXBuffer_: array[0..1023] of Byte;
        fRXBufferLength_:integer;
        fRXBufferTimeOut:integer;

        fPageBuffer_:array[0..1023] of Byte;
        fPageBufferLength_:integer;
        fPageBufferAddress_:integer;
        fFileType_:TFileType;
        fBaud:integer;
        fSerPath:string;

        fBaudUsed:integer;
        fSerPathUsed:string;

        fErrorFlag_:integer;
        FOnError:  TErrorEvent;
        FOnStatus: TStatusEvent;
        FOnRXByte: TRXEvent;
        FOnNewPage: TNewPageEvent;
        FOnBootInfo: TBootInfoEvent;

        function CheckCMDOK: boolean;
        procedure EventNewPageFileType;
        procedure EventStatus;
        procedure EventError;
        procedure EventRX;
        procedure EventBootInfo;
        procedure EventNewPage;

        //function CRC16(crc:word; data: array of byte; length:integer):word;
        procedure OnStatus_(Sender: TObject; Reason: THookSerialReason;
          const Value: string);


        procedure SyncTimerStart;
        procedure SyncTimerElapsed;
        procedure SyncState;
        procedure SyncBaud;
        procedure SyncSerPath;
        procedure SyncSerTX;
        procedure SyncSerRX;
        function ScanForBootInfo:boolean;
        procedure sendbuffer(Buffer:array of Byte;length:integer);

        function CheckCMDNeedsAck(var Data: array of Byte; Length: integer):boolean;
        procedure SendCMDCustom(var Data: array of Byte; Length: integer);
        procedure SendCMDCatchString;
        procedure SendCMDKeppAlive;
        procedure SendCMDErasePages(PageCount: Byte);
        procedure SendCMDFillBuffer(var Data:array of Byte; Length:integer);
        procedure SendCMDStartAddress(StartAddress: integer);
        procedure SendCMDWritePage;
        procedure SendCMDRun;
      public

        property OnStatus: TStatusEvent read FOnStatus write FOnStatus;
        property OnError: TErrorEvent read FOnError write FOnError;
        property OnRxByte: TRXEvent read FOnRXByte write FOnRXByte;
        property OnBootInfo: TBootInfoEvent read FOnBootInfo write FOnBootInfo;
        property OnNewPage: TNewPageEvent read FOnNewPage write FOnNewPage;

        constructor create;
        destructor destroy; override;
        procedure GoProgramming;
        procedure GoConnect;
        procedure GoDisconnect;
        procedure SetBaud(baud:integer);
        procedure SetSerPath(Path:string);
        function GetBaud:integer;
        function GetSerPath:string;
        function GetStatus:TBootloaderStatus;
      protected
        procedure Execute; override;

    end;

    TBootloader = class
      private
        Thread:TBootloaderThread;
        ThreadTerminated:boolean;
        InFile:TInputFile;
        EPikTimer:TEpikTimer;
        fTimerData:TimerData;
        fBootloaderStatusBeforeCMD:TBootloaderStatus;
        //BootInfo:TBootInfo;

        FOnBootInfo: TBootInfoEvent;
        FOnStatus: TStatusStringEvent;
        FOnRXByte: TRXEvent;
        FOnProgramProgress: TBootloaderProgressEvent;
        FOnError: TErrorStringEvent;

        procedure OnBootloaderError(Error: TBootloaderError;
          Status: TBootloaderStatus; CMDStatus: TBootloaderCMDStatus);
        procedure OnBootloaderStatus(Status: TBootloaderStatus; CMDStatus: TBootloaderCMDStatus);
        procedure OnBootloaderRXByte(RxByte: Byte);
        procedure OnBootloaderBootInfo(BootInfo: TBootInfo);
        procedure OnNewPage(var address:integer;var Buffer: array of Byte; var Length:integer; const PageSize:integer;var FileType:TFileType);
        procedure OnThreadTerminate(Sender: TObject);
      public
        property OnError: TErrorStringEvent read FOnError write FOnError;
        property OnStatus: TStatusStringEvent read FOnStatus write FOnStatus;
        property OnRxByte: TRXEvent read FOnRXByte write FOnRXByte;
        property OnBootInfo: TBootInfoEvent read FOnBootInfo write FOnBootInfo;
        property OnProgramProgress: TBootloaderProgressEvent read FOnProgramProgress write FOnProgramProgress;

        constructor create;
        destructor destroy;  override;
        procedure Connect;
        procedure ProgramFlash(path:string);

        procedure Disconnect;
        procedure SetTTY(path:string);
        procedure SetBaud(Baud:integer);

        function GetBaud:integer;
        function GetTTY:string;

        function GetStatus:TBootloaderStatus;
    end;
implementation

{ TBootloader }

function CRC16(crc: word; var data: array of byte; length: integer
  ): word;
var
  poly:word;
  index,i:integer;
begin
  poly := $A001;
  index := 0;
  while (length > 0) do begin
    dec(length);
    crc := crc xor data[index];
    inc(index);
    for i := 8 downto 1 do begin
      if (crc and 1) = 1 then begin
        crc := crc shr 1;
        crc := crc xor poly;
      end else begin
          crc := crc shr 1;
      end;
    end;
  end;
  result := crc;
end;

constructor TBootloaderThread.create;
begin
  FreeOnTerminate := false;
  inherited create(true);
  fBaud:=0;
  fSerPath:='';
  fBaudUsed:=0;
  fSerPathUsed:='';
  fBootloaderCommand:=bcIDLE;
  fTXBufferLength_:=0;
  fRXBufferLength_:=0;
  fRXBufferTimeOut:=10;

  ser:=TBlockserial.Create;
  ser.OnStatus:= @OnStatus_;
  ser.LinuxLock:=false;
  fEpikTimer := TEpikTimer.Create(nil);

end;

destructor TBootloaderThread.destroy;
begin
  fEpikTimer.free;
  ser.Free;
  inherited destroy;
end;

procedure TBootloaderThread.OnStatus_(Sender: TObject; Reason: THookSerialReason;
    const Value: string);

begin
  if Reason = HR_CanRead then begin
     reason := HR_CanRead;
  end;
  if Reason = HR_ReadCount then begin
     reason := HR_ReadCount;
  end;
end;

procedure TBootloaderThread.SyncTimerStart;
begin
  fEpikTimer.Clear(fTimerData);
  fEpikTimer.Start(fTimerData);
end;

procedure TBootloaderThread.SyncTimerElapsed;
begin
  fTimerElapsed_:=fEpikTimer.Elapsed(fTimerData);
end;



procedure TBootloaderThread.sendbuffer(Buffer: array of Byte; length: integer);
var
  i:integer;
  b:byte;
  crc:word;
begin
  for i := 0 to length-1 do begin
    b := buffer[i];
    ser.SendByte(b);
  end;
  crc := CRC16(0,buffer,length);
  ser.SendByte(crc);
  ser.SendByte(crc shr 8);
end;





procedure TBootloaderThread.SendCMDCatchString;
begin
    fTXBuffer_[0]:=($00);
    fTXBuffer_[1]:=($00);
    fTXBuffer_[2]:=($00);
    fTXBuffer_[3]:=($00);

    fTXBuffer_[4]:=($00);
    fTXBuffer_[5]:=($00);
    fTXBuffer_[6]:=($00);
    fTXBuffer_[7]:=($00);
    fTXBuffer_[8]:=($00);

    fTXBuffer_[9]:=($00);
    fTXBuffer_[10]:=($42);
    fTXBuffer_[11]:=($4F);
    fTXBuffer_[12]:=($4F);
    fTXBuffer_[13]:=($54);
    fTXBuffer_[14]:=($4C);
    fTXBuffer_[15]:=($4F);
    fTXBuffer_[16]:=($41);
    fTXBuffer_[17]:=($44);

    fTXBuffer_[18]:=($45);
    fTXBuffer_[19]:=($52);
    fTXBufferLength_:= 20;
    Synchronize(@SyncSerTX);

end;

procedure TBootloaderThread.SendCMDKeppAlive;
begin
    fTXBuffer_[0]:=$FD;
    fTXBuffer_[1]:=$00;
    fTXBufferLength_:= 2;
    Synchronize(@SyncSerTX);
end;

procedure TBootloaderThread.SendCMDRun;
begin
    fTXBuffer_[0]:=$00;
    fTXBuffer_[1]:=$01;
    fTXBufferLength_:= 2;
    Synchronize(@SyncSerTX);
end;

procedure TBootloaderThread.SendCMDStartAddress(StartAddress:integer);
begin
    fTXBuffer_[0]:=$FF;
    fTXBuffer_[1]:=StartAddress shr 16;
    fTXBuffer_[2]:=StartAddress shr 8;
    fTXBuffer_[3]:=StartAddress;
    fTXBufferLength_:= 4;

    Synchronize(@SyncSerTX);
end;

procedure TBootloaderThread.SendCMDFillBuffer(var Data:array of Byte; Length:integer);
var
  i:integer;
begin
    fTXBuffer_[0]:=$FE;
    fTXBuffer_[1]:=$00;
    fTXBuffer_[2]:=Length shr 8;
    fTXBuffer_[3]:=Length;
    fTXBufferLength_:= 4;
    Synchronize(@SyncSerTX);

    for i := 0 to Length-1 do begin
      fTXBuffer_[i]:= Data[i];
    end;
    fTXBufferLength_:= Length;
    Synchronize(@SyncSerTX);
end;



procedure TBootloaderThread.SendCMDWritePage;
begin
    fTXBuffer_[0]:=($01);
    fTXBuffer_[1]:=($01);
    fTXBufferLength_:= 2;
    Synchronize(@SyncSerTX);
end;

procedure TBootloaderThread.SendCMDErasePages(PageCount:Byte);
begin
    fTXBuffer_[0]:=($02);
    fTXBuffer_[1]:=(PageCount);
    fTXBufferLength_:= 2;
    Synchronize(@SyncSerTX);
end;

procedure TBootloaderThread.SendCMDCustom(var Data:array of Byte; Length:integer);
var
  i:integer;
begin
    for i := 0 to Length-1 do begin
      fTXBuffer_[i]:= Data[i];
    end;
    fTXBufferLength_:= Length;
    Synchronize(@SyncSerTX);
end;

function TBootloaderThread.CheckCMDNeedsAck(var Data: array of Byte;
  Length: integer): boolean;
begin
  result := data[0] <> $FE;
end;

function TBootloaderThread.ScanForBootInfo: boolean;
var
  i,n:integer;
  c:boolean;
begin
 // Synchronize(@SyncSerRX);
  if fRXBufferLength_ > 0 then begin
    result := false;
    for n := fRXBufferLength_-1 downto 0 do begin    //Lets look for the 'OK' Message (='0')
      if (fRXBuffer_[n] = ord(beSuccessPlainStyle)) or (fRXBuffer_[n] = ord(beSuccessCryptStyle)) then begin
        result := true;
        break;
      end;
    end;
    if result then begin
      BootInfo_.Crypt:=fRXBuffer_[n] = ord(beSuccessCryptStyle);
      BootInfo_.Bootpages:=fRXBuffer_[n-1];
      BootInfo_.version:=fRXBuffer_[n-2];
      BootInfo_.Sign2:=fRXBuffer_[n-3];
      BootInfo_.Sign1:=fRXBuffer_[n-4];
      BootInfo_.Sign0:=$1E;
      BootInfo_.Hello:='';
      for i := 0 to n-5 do begin
        if (fRXBuffer_[i] = ord(beSuccessPlainStyle)) or (fRXBuffer_[i] = ord(beSuccessCryptStyle)) then
          break;
        BootInfo_.Hello:= BootInfo_.Hello+char(fRXBuffer_[i]);
      end;
      if (BootInfo_.Sign2 = $07) and (BootInfo_.Sign1=$93) then begin
        BootInfo_.FlashSize:=1024*8;
      end else
        BootInfo_.FlashSize :=-1;
    end;
  end else
    result := false;
end;

function TBootloaderThread.CheckCMDOK: boolean;
var
  b:byte;
begin
  if fRXBufferLength_ > 0 then begin
    b := fRXBuffer_[fRXBufferLength_-1];
    result := (b = ord(beSuccessPlainStyle)) or (b = ord(beSuccessCryptStyle));
   // if not Result then
      fErrorFlag_ := b;
  end else begin
    result := false;
  end;
end;


procedure TBootloaderThread.EventStatus;
begin
if Assigned(FOnStatus) then
  begin
    FOnStatus(fBootloaderIsStatus_,fBootloaderCommandStatus_);
  end;
end;

procedure TBootloaderThread.EventError;
begin
if Assigned(FOnError) then
  begin
    FOnError(TBootloaderError(fErrorFlag_),fBootloaderIsStatus_,fBootloaderCommandStatus_);
  end;
end;

procedure TBootloaderThread.EventRX;
begin
if Assigned(FOnRXByte) then
  begin
    FOnRXByte(fRXBuffer_[fRXBufferLength_]);
  end;
end;

procedure TBootloaderThread.EventBootInfo;
begin
if Assigned(FOnBootInfo) then
  begin
    FOnBootInfo(BootInfo_);
  end;
end;

procedure TBootloaderThread.EventNewPage;
var
  i:integer;
begin
if Assigned(FOnNewPage) then
  begin
    if fFileType_ = ftAcy then
      i := high(fPageBuffer_)+1
    else
      i := 960;
    FOnNewPage(fPageBufferAddress_,fPageBuffer_,fPageBufferLength_,i,fFileType_);
  end;
end;

procedure TBootloaderThread.EventNewPageFileType;
var
  i:integer;
begin
if Assigned(FOnNewPage) then
  begin
    i := 0;
    FOnNewPage(fPageBufferAddress_,fPageBuffer_,fPageBufferLength_,i,fFileType_);
  end;
end;

procedure TBootloaderThread.GoProgramming;
begin
  fBootloaderCommand:= bcProgram;
end;

procedure TBootloaderThread.GoConnect;
begin
  fBootloaderCommand:= bcKeepAlive;
end;

procedure TBootloaderThread.GoDisconnect;
begin
  fBootloaderCommand:= bcIDLE;
end;

procedure TBootloaderThread.SetBaud(baud: integer);
begin
  fBaud:=baud;
  fBaudUsed:=baud;
end;

procedure TBootloaderThread.SetSerPath(Path: string);
begin
  fSerPath:= Path;
  fSerPathUsed:=Path;
end;

function TBootloaderThread.GetBaud: integer;
begin
  result := fBaudUsed;
end;

function TBootloaderThread.GetSerPath: string;
begin
  result := fSerPathUsed;
end;

function TBootloaderThread.GetStatus: TBootloaderStatus;
begin
  result := fBootloaderIsStatus;
end;

procedure TBootloaderThread.SyncState;
begin
  fBootloaderIsStatus := fBootloaderIsStatus_;
  if fBootloaderCommand <> bcNone then begin
    if fBootloaderCommand=bcIDLE then
      fBootloaderShouldStatus_ := bsIDLE
    else
      fBootloaderShouldStatus_:= bsConnected;

    fBootloaderCommand_:= fBootloaderCommand;
    fFileType_ := ftnone;
    fBootloaderCommand:= bcNone;
  end;
end;

procedure TBootloaderThread.SyncBaud;
begin
  if fBaud <> 0 then begin
    ser.Config(fBaud,8,'N',0,false,false);
    fBaud := 0;
  end;
end;

procedure TBootloaderThread.SyncSerPath;
begin
  if fSerPath <> '' then begin
    ser.free;
    ser := TBlockSerial.Create;
    ser.LinuxLock:=false;
    ser.Connect(fSerPath);
    fBaud:=fBaudUsed;
    fSerPath := '';
  end;
end;

procedure TBootloaderThread.SyncSerTX;
begin
  sendbuffer(fTXBuffer_, fTXBufferLength_);
end;

procedure TBootloaderThread.SyncSerRX;
begin
  fRXBufferLength_:=0;
  while ser.CanReadEx(fRXBufferTimeOut) do begin
    fRXBuffer_[fRXBufferLength_] := ser.RecvByte(fRXBufferTimeOut);
    EventRX;
    inc(fRXBufferLength_);
  end;
end;

procedure TBootloaderThread.Execute;
var
  ScanResult:boolean;
  OldIsStyle:TBootloaderStatus;
  OldCMDStatus:TBootloaderCMDStatus;
  CMDStatusmask:integer;
  ProgramLength:integer;
  ProgramStartAddress:integer;
  PagesToErase:integer;

  Timeout:extended;
begin

  try
    //EpikTimer.Start(Timerd);
    fBootloaderShouldStatus_ := bsIDLE;
    fBootloaderIsStatus_ := bsIDLE;
    OldIsStyle := fBootloaderIsStatus_;
    OldCMDStatus := fBootloaderCommandStatus_;
    ProgramLength := 0;
    Synchronize(@SyncTimerStart);
    Timeout:=0;
    repeat
      Synchronize(@SyncState);
      if fBootloaderIsStatus_ <> bsIDLE then
        Synchronize(@SyncSerRX);
      fErrorFlag_ := 0;
      if (fBootloaderShouldStatus_ <> fBootloaderIsStatus_) then begin
        if (fBootloaderShouldStatus_ > fBootloaderIsStatus_)then begin
          if (fBootloaderIsStatus_ = bsIDLE) or (fBootloaderIsStatus_ = bsConnecting) then begin
            if fBootloaderIsStatus_ = bsIDLE then
              fBootloaderIsStatus_ := bsConnecting;
            fBootloaderCommandStatus_ :=csIDlE;
            SendCMDCatchString;
            ScanResult := ScanForBootInfo;
            if ScanResult then begin
              fBootloaderIsStatus_ := bsConnected;
              Synchronize(@EventBootInfo);
            end;

          end;
        end else begin
          Sleep(200);
          SendCMDRun;
          fBootloaderIsStatus_ := bsIDLE;
          fBootloaderCommandStatus_ :=csIDlE;
        end;
      end else begin

        if fBootloaderIsStatus_ = bsConnected then begin
          CMDStatusmask:=ord(fBootloaderCommandStatus_);
          case fBootloaderCommand_ of
            bcKeepAlive:  begin
                            fBootloaderCommandStatus_:=csKeepAlive;
                            SendCMDKeppAlive;
                            Timeout := 0;
                          end;
            bcProgram:    begin
                            if (CMDStatusmask and $F0) <> (ord(csProg_AddressSetted) and $F0) then begin //We just started the Programmingmode
                              if BootInfo_.Crypt then begin
                                fBootloaderCommandStatus_ := csCrypProg_Start;
                              end else begin
                                Synchronize(@EventNewPage);
                                if fFileType_ = ftacy then begin
                                  fErrorFlag_ := ord(beCrypt);
                                end else begin
                                  SendCMDStartAddress(fPageBufferAddress_);
                                  ProgramStartAddress:= fPageBufferAddress_;                                //Set Start Address
                                  ProgramLength := fPageBufferLength_;
                                  fBootloaderCommandStatus_:=csProg_AddressSetted;
                                  Synchronize(@SyncTimerStart);
                                  Timeout := 500/1000;
                                end;
                              end;
                            end else if fBootloaderCommandStatus_ = csProg_AddressSetted then begin
                              if CheckCMDOK then
                                fBootloaderCommandStatus_ := csProg_AddressOK;
                            end else if (fBootloaderCommandStatus_ = csProg_AddressOK) or
                                        (fBootloaderCommandStatus_ = csProg_StartNextBuffer) then begin
                              SendCMDFillBuffer(fPageBuffer_,fPageBufferLength_);                         //Fill Buffer
                              fBootloaderCommandStatus_:=csProg_BufferFilled;
                              Synchronize(@SyncTimerStart);
                              Timeout := 1000/1000;
                            end else if fBootloaderCommandStatus_ = csProg_BufferFilled then begin
                              if CheckCMDOK then begin
                                fBootloaderCommandStatus_ := csProg_BufferOK;
                              end;
                            end else if fBootloaderCommandStatus_ = csProg_BufferOK then begin
                              SendCMDWritePage;                                                           //Write Buffer
                              fBootloaderCommandStatus_:=csProg_WritePageSent;
                              Synchronize(@SyncTimerStart);
                              Timeout := 500/1000;
                            end else if fBootloaderCommandStatus_ = csProg_WritePageSent then begin
                              if CheckCMDOK then begin
                                fBootloaderCommandStatus_ := csProg_WritePageOK;
                              end;
                            end else if fBootloaderCommandStatus_ = csProg_WritePageOK then begin
                              Synchronize(@EventNewPage);
                              ProgramLength := ProgramLength + fPageBufferLength_;
                              if fPageBufferLength_<> 0 then
                                fBootloaderCommandStatus_:=csProg_StartNextBuffer
                              else begin
                                PagesToErase:= BootInfo_.FlashSize;
                                PagesToErase:= PagesToErase-ProgramLength;
                                PagesToErase:= PagesToErase-ProgramStartAddress;
                                PagesToErase:= PagesToErase div PAGE_SIZE;
                                PagesToErase:= PagesToErase-BootInfo_.Bootpages;
                                SendCMDErasePages(PagesToErase);
                                fBootloaderCommandStatus_:=csProg_EraseSent;
                                Synchronize(@SyncTimerStart);
                                Timeout := 500/1000;
                              end;

                            end else if fBootloaderCommandStatus_ = csProg_EraseSent then begin
                              if CheckCMDOK then begin
                                fBootloaderCommandStatus_ := csProg_EraseOK;
                              end;
                            end else if fBootloaderCommandStatus_ = csProg_EraseOK then begin
                              fBootloaderCommandStatus_ := csProg_Ready;
                              Timeout := 0;
                            end  else if (fBootloaderCommandStatus_ = csCrypProg_Start) or (fBootloaderCommandStatus_ = csCrypProg_StartNextBuffer) then begin
                              Synchronize(@EventNewPage);
                              if fFileType_ = ftAcy then begin
                                fBootloaderCommandStatus_ := csCryptProg_CMDSent;
                                if fPageBufferLength_ = 0 then begin
                                  fBootloaderCommandStatus_ := csProg_Ready;
                                end else begin
                                  SendCMDCustom(fPageBuffer_,fPageBufferLength_);
                                  if not CheckCMDNeedsAck(fPageBuffer_,fPageBufferLength_) then begin
                                    Synchronize(@EventNewPage);
                                    if fPageBufferLength_ = 0 then begin
                                      fBootloaderCommandStatus_ := csProg_Ready
                                    end else
                                      SendCMDCustom(fPageBuffer_,fPageBufferLength_);
                                  end;
                                end;
                                Synchronize(@SyncTimerStart);
                                Timeout := 5000/1000;
                              end else begin
                                  fErrorFlag_ := ord(beCrypt);
                              end;


                            end else if fBootloaderCommandStatus_ = csCryptProg_CMDSent then begin
                              if CheckCMDOK then begin
                                fBootloaderCommandStatus_ := csCrypProg_StartNextBuffer;
                              end;
                            end else if (fBootloaderCommandStatus_ = csProg_Ready) or (fBootloaderCommandStatus_ = csProgError) then begin
                              fBootloaderCommand_ := bcKeepAlive;
                              Timeout := 0;
                            end

                          end;
          end;

        end;
      end;
      if fBootloaderCommand_ = bcProgram then begin
        if (fErrorFlag_ <> 0) then begin
          Synchronize(@EventError);
          if (fErrorFlag_ <> ord(beSuccessCryptStyle)) and (fErrorFlag_ <> ord(beSuccessPlainStyle)) then begin
            fBootloaderCommandStatus_ := csProgError;
          end else begin
           // Timeout:= 0;
          end;
        end;
        Synchronize(@SyncTimerElapsed);
        if (fTimerElapsed_>Timeout)and (Timeout <> 0) then begin
          Timeout := 0;
          fErrorFlag_:=ord(beTimeOut);
          Synchronize(@EventError);
        end;
        sleep(1);
      end else
        sleep(200);

      if (OldIsStyle <> fBootloaderIsStatus_) or (OldCMDStatus <> fBootloaderCommandStatus_) then
        Synchronize(@EventStatus);
      OldIsStyle:=fBootloaderIsStatus_;
      OldCMDStatus:=fBootloaderCommandStatus_;


      Synchronize(@SyncSerPath);
      Synchronize(@SyncBaud);
    until Terminated;
  finally

  end;
end;

{ TBootloader }

constructor TBootloader.create;
begin
  inherited create;
  InFile := TInputFile.Create;
  EPikTimer := TEpikTimer.create(nil);
  ThreadTerminated:= false;
  Thread := TBootloaderThread.Create;
  Thread.OnStatus:= @OnBootloaderStatus;
  Thread.OnRxByte:= @OnBootloaderRXByte;
  Thread.OnBootInfo:= @OnBootloaderBootInfo;
  Thread.OnNewPage := @OnNewPage;
  Thread.OnTerminate:=@OnThreadTerminate;
  Thread.OnError := @OnBootloaderError;
  Thread.Resume;
end;

destructor TBootloader.destroy;
begin
  Thread.Terminate;

  repeat
    Application.ProcessMessages;
  until ThreadTerminated;
  Thread.Free;
  EpikTimer.free;
  inherited destroy;
end;

procedure TBootloader.Connect;
begin
  Thread.GoConnect;
end;

procedure TBootloader.OnNewPage(var address: integer;
  var Buffer: array of Byte; var Length: integer; const PageSize: integer; var FileType: TFileType);
begin
  address:= InFile.GetNextAddress;
  Length := InFile.ReadContinue(Buffer,PageSize);
  FileType := InFile.GetFileType;
end;

procedure TBootloader.ProgramFlash(path: string);

  function CheckCMDNeedsAck(var Data: array of Byte): boolean;
  begin
    result := data[0] <> $FE;
  end;

var
  test:TStringList;
  buf:array[0..1025]of byte;
  i:integer;
  l:integer;
  crc:word;
begin
  fBootloaderStatusBeforeCMD := Thread.GetStatus;
  InFile.LoadFromFile(Path);
  InFile.SetPadding(64,255);

 { test := TStringList.Create;
  try
    InFile.GotoAddress(960);
    repeat
      l := InFile.ReadContinue(buf,512);
      for i := 0 to l-1 do begin
        if i mod 16 = 0 then
          test.add('');
        test[test.count-1] := test[test.count-1] + inttohex(buf[i],2)+' ';
      end;
    until l = 0;
    test.SaveToFile('/home/arne/Desktop/hex.txt');
  finally
    test.Free;
  end;     }

 { if InFile.GetFileType=ftAcy then begin
    test := TStringList.Create;
    try
      repeat
        l := InFile.ReadContinue(buf,1024);
        if l <> 0 then begin
          crc := CRC16(0,buf,l);
          buf[l] := crc;
          buf[l+1] := crc shr 8;
          for i := 0 to l+1 do begin
            if i mod 16 = 0 then
              test.add('');
            test[test.count-1] := test[test.count-1] + inttohex(buf[i],2)+' ';
          end;

          if CheckCMDNeedsAck(buf[0]) then
            test.add('7')
          else
            test.add('');
        end;
      until l = 0;
      test.SaveToFile('/home/arne/Desktop/hex.txt');
    finally
      test.Free;
    end;
  end;  }
  // else
    Thread.GoProgramming;
end;

procedure TBootloader.Disconnect;
begin
  Thread.GoDisconnect;
end;

procedure TBootloader.SetTTY(path: string);
begin
  Thread.SetSerPath(path);
end;

procedure TBootloader.SetBaud(Baud: integer);
begin
  Thread.SetBaud(Baud);
end;

function TBootloader.GetBaud: integer;
begin
  result := Thread.GetBaud;
end;

function TBootloader.GetTTY: string;
begin
  result := Thread.GetSerPath;
end;

function TBootloader.GetStatus: TBootloaderStatus;
begin
  result := Thread.GetStatus;
end;


procedure TBootloader.OnBootloaderError(Error: TBootloaderError; Status: TBootloaderStatus; CMDStatus: TBootloaderCMDStatus);
var
  s:string;
begin
  if Assigned(FOnError) then begin
    case Error of
      beSuccessPlainStyle:  s := 'OK';
      beSuccessCryptStyle:  s := 'OK';
      beBounds:             s := 'Bounds error';
      beCommand:            s := 'Command error';
      beCRC:                s := 'CRC error';
      beCrypt:              s := 'Encryption error';
      beProg:               s := 'Programming error';
      beVerify:             s := 'Verify error';
      beVersion:            s := 'Version error';
      beTimeout:            s := 'Timeout error';
    end;
    if (Error <> beSuccessPlainStyle) and (Error <> beSuccessCryptStyle) and (Error <> beNone) then begin
      s := s+' (0x'+inttohex(ord(Error),2)+')';
    end;
    FOnError(s,Error);
  end;
end;

procedure TBootloader.OnBootloaderStatus(Status: TBootloaderStatus;CMDStatus: TBootloaderCMDStatus);
var
  s:string;
  nl:boolean;
  fs:TFormatSettings;
begin
  if Assigned(FOnStatus) then begin
    nl := true;
    case Status of
      bsIDLE: s := 'IDLE..';
      bsConnected:  begin

                      case CMDStatus of
                        csKeepAlive:            s := 'Keeping Connection..';
                        csProg_AddressSetted:   begin
                                                  EPikTimer.Clear(fTimerData);
                                                  EPikTimer.Start(fTimerData);
                                                  s := 'Starting Programming @ Address 0x'+InttoHex(InFile.GetStartAddress,4)+'..';
                                                  nl := false;
                                                end;
                        csCrypProg_Start:       begin
                                                  EPikTimer.Clear(fTimerData);
                                                  EPikTimer.Start(fTimerData);
                                                  s := 'Starting Programming @ Address 0x'+InttoHex(InFile.GetStartAddress,4)+'..';
                                                  nl := true;
                                                end;
                        //csProg_AddressOK:       s := 'Programming started';
                        csProg_StartNextBuffer: s := 'Next Buffer..';
                        csProg_BufferFilled:    begin s := #9+'Filling buffer with 0x'+InttoHex(InFile.GetLatestBufferSize,4)+' bytes..'; nl := false; end;
                        //csProg_BufferOK:        s := #9+'Buffer filled';
                        csProg_WritePageSent:   begin s := #9+'Writing buffer..'; nl := false; end;
                        //csProg_WritePageOK:     s := #9+'Buffer written';
                        csProg_EraseSent:       begin s := #9+'Erasing empty flash..'; nl := false; end;
                        //csProg_EraseOK:         s := #9+'Empty flash erased';
                        csProg_Ready:           begin

                                                  fs := DefaultFormatSettings;
                                                  //fs.
                                                  s :=  'Device successfully programmed with '+InttoStr(InFile.GetTotalTranferedBytesTotal)+
                                                        ' bytes in '+CurrToStr(EPikTimer.Elapsed(fTimerData))+' sec!';
                                                  EPikTimer.Stop(fTimerData);
                                                end;
                        csProgError:            begin

                                                  fs := DefaultFormatSettings;
                                                  //fs.
                                                  s :=  'Programming stopped';
                                                  EPikTimer.Stop(fTimerData);
                                                end;

                        csCryptProg_CMDSent:    begin
                                                  if InFile.GetLatestBufferSize > 20 then
                                                    s := #9+'Filling buffer with 0x'+InttoHex(InFile.GetLatestBufferSize,4)+' bytes..'
                                                  else
                                                    s := #9+'Sending Bootloadercommand..';
                                                  nl := false;
                                                end;
                      end;
                    end;

      bsConnecting: s := 'Connecting..';
    end;
    FOnStatus(s,status,nl);
    if (CMDStatus = csProg_Ready) and (fBootloaderStatusBeforeCMD = bsIDLE) then begin
      Disconnect;
    end;
  end;
end;


procedure TBootloader.OnBootloaderRXByte(RxByte: Byte);
var
  s:string;
begin
  if Assigned(FOnRXByte) then begin
    FOnRXByte(RxByte);
  end;
end;

procedure TBootloader.OnBootloaderBootInfo(BootInfo: TBootInfo);
begin
  //BootInfo :=
  if Assigned(FOnBootInfo) then begin
    FOnBootInfo(BootInfo);
  end;
end;



procedure TBootloader.OnThreadTerminate(Sender: TObject);
begin
  ThreadTerminated:=true;
end;


end.

